@@ -65,6 +65,7 @@ struct _UDisksClient
 typedef struct
 {
   GObjectClass parent_class;
+  GHashTable *changed_blacklist;
 } UDisksClientClass;
 
 enum
@@ -131,24 +132,28 @@ udisks_client_finalize (GObject *object)
     g_source_destroy (client->changed_timeout_source);
 
   if (client->initialization_error != NULL)
-    g_error_free (client->initialization_error);
+    g_clear_error (&(client->initialization_error));
 
-  g_signal_handlers_disconnect_by_func (client->object_manager,
-                                        G_CALLBACK (on_object_added),
-                                        client);
-  g_signal_handlers_disconnect_by_func (client->object_manager,
-                                        G_CALLBACK (on_object_removed),
-                                        client);
-  g_signal_handlers_disconnect_by_func (client->object_manager,
-                                        G_CALLBACK (on_interface_added),
-                                        client);
-  g_signal_handlers_disconnect_by_func (client->object_manager,
-                                        G_CALLBACK (on_interface_removed),
-                                        client);
-  g_signal_handlers_disconnect_by_func (client->object_manager,
-                                        G_CALLBACK (on_interface_proxy_properties_changed),
-                                        client);
-  g_object_unref (client->object_manager);
+  /* might be NULL if failing early in the constructor */
+  if (client->object_manager != NULL)
+    {
+      g_signal_handlers_disconnect_by_func (client->object_manager,
+                                            G_CALLBACK (on_object_added),
+                                            client);
+      g_signal_handlers_disconnect_by_func (client->object_manager,
+                                            G_CALLBACK (on_object_removed),
+                                            client);
+      g_signal_handlers_disconnect_by_func (client->object_manager,
+                                            G_CALLBACK (on_interface_added),
+                                            client);
+      g_signal_handlers_disconnect_by_func (client->object_manager,
+                                            G_CALLBACK (on_interface_removed),
+                                            client);
+      g_signal_handlers_disconnect_by_func (client->object_manager,
+                                            G_CALLBACK (on_interface_proxy_properties_changed),
+                                            client);
+      g_object_unref (client->object_manager);
+    }
 
   if (client->context != NULL)
     g_main_context_unref (client->context);
@@ -201,6 +206,11 @@ udisks_client_class_init (UDisksClientClass *klass)
   gobject_class->finalize     = udisks_client_finalize;
   gobject_class->get_property = udisks_client_get_property;
 
+  klass->changed_blacklist = g_hash_table_new (g_str_hash, g_str_equal);
+  g_hash_table_insert (klass->changed_blacklist, (gpointer) "SmartSelftestPercentRemaining", NULL);
+  g_hash_table_insert (klass->changed_blacklist, (gpointer) "SyncRate", NULL);
+  g_hash_table_insert (klass->changed_blacklist, (gpointer) "SyncRemainingTime", NULL);
+
   /**
    * UDisksClient:object-manager:
    *
@@ -387,11 +397,9 @@ initable_init (GInitable     *initable,
         {
           init_interface_proxy (client, G_DBUS_PROXY (ll->data));
         }
-      g_list_foreach (interfaces, (GFunc) g_object_unref, NULL);
-      g_list_free (interfaces);
+      g_list_free_full (interfaces, g_object_unref);
     }
-  g_list_foreach (objects, (GFunc) g_object_unref, NULL);
-  g_list_free (objects);
+  g_list_free_full (objects, g_object_unref);
 
   g_signal_connect (client->object_manager,
                     "object-added",
@@ -593,8 +601,7 @@ udisks_client_get_block_for_label (UDisksClient        *client,
         g_object_unref (block);
     }
 
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
+  g_list_free_full (object_proxies, g_object_unref);
   ret = g_list_reverse (ret);
   return ret;
 }
@@ -638,8 +645,7 @@ udisks_client_get_block_for_uuid (UDisksClient        *client,
         g_object_unref (block);
     }
 
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
+  g_list_free_full (object_proxies, g_object_unref);
   ret = g_list_reverse (ret);
   return ret;
 }
@@ -686,8 +692,7 @@ udisks_client_get_block_for_dev (UDisksClient *client,
     }
 
  out:
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
+  g_list_free_full (object_proxies, g_object_unref);
   return ret;
 }
 
@@ -724,8 +729,7 @@ get_top_level_blocks_for_drive (UDisksClient *client,
         }
       g_object_unref (block);
     }
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
+  g_list_free_full (object_proxies, g_object_unref);
   return ret;
 }
 
@@ -779,8 +783,7 @@ udisks_client_get_block_for_drive (UDisksClient        *client,
     }
 
  out:
-  g_list_foreach (blocks, (GFunc) g_object_unref, NULL);
-  g_list_free (blocks);
+  g_list_free_full (blocks, g_object_unref);
   return ret;
 }
 
@@ -844,6 +847,65 @@ udisks_client_get_mdraid_for_block (UDisksClient  *client,
 
 /* ---------------------------------------------------------------------------------------------------- */
 
+typedef const gchar *
+(*member_retrieve_cb) (UDisksBlock *object);
+
+static GList *
+_udisks_client_get_block_or_blocks_for_mdraid (UDisksClient *client,
+                                               UDisksMDRaid *raid,
+                                               member_retrieve_cb member_get,
+                                               gboolean only_first_one,
+                                               gboolean skip_partitions)
+{
+  GList *ret = NULL;
+  GList *l, *object_proxies = NULL;
+  GDBusObject *raid_object;
+  const gchar *raid_objpath;
+
+  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
+  g_return_val_if_fail (UDISKS_IS_MDRAID (raid), NULL);
+
+  raid_object = g_dbus_interface_get_object (G_DBUS_INTERFACE (raid));
+  if (raid_object == NULL)
+    goto out;
+
+  raid_objpath = g_dbus_object_get_object_path (raid_object);
+
+  object_proxies = g_dbus_object_manager_get_objects (client->object_manager);
+  for (l = object_proxies; l != NULL; l = l->next)
+    {
+      UDisksObject *object = UDISKS_OBJECT (l->data);
+      UDisksBlock *block;
+
+      block = udisks_object_get_block (object);
+      if (block == NULL)
+        continue;
+
+      /* ignore partitions */
+      if (skip_partitions)
+        {
+          if (udisks_object_peek_partition (object) != NULL)
+            continue;
+        }
+
+      if (g_strcmp0 (member_get (block), raid_objpath) == 0)
+        {
+          ret = g_list_prepend (ret, block);
+
+          if (only_first_one)
+            goto out;
+        }
+      else
+        {
+          g_object_unref (block);
+        }
+    }
+
+ out:
+  g_list_free_full (object_proxies, g_object_unref);
+  return ret;
+}
+
 /**
  * udisks_client_get_block_for_mdraid:
  * @client: A #UDisksClient.
@@ -869,45 +931,19 @@ UDisksBlock *
 udisks_client_get_block_for_mdraid (UDisksClient *client,
                                     UDisksMDRaid *raid)
 {
+  GList *b_list = NULL;
   UDisksBlock *ret = NULL;
-  GList *l, *object_proxies = NULL;
-  GDBusObject *raid_object;
-  const gchar *raid_objpath;
 
-  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
-  g_return_val_if_fail (UDISKS_IS_MDRAID (raid), NULL);
-
-  raid_object = g_dbus_interface_get_object (G_DBUS_INTERFACE (raid));
-  if (raid_object == NULL)
-    goto out;
-
-  raid_objpath = g_dbus_object_get_object_path (raid_object);
-
-  object_proxies = g_dbus_object_manager_get_objects (client->object_manager);
-  for (l = object_proxies; l != NULL; l = l->next)
-    {
-      UDisksObject *object = UDISKS_OBJECT (l->data);
-      UDisksBlock *block;
-
-      block = udisks_object_get_block (object);
-      if (block == NULL)
-        continue;
-
-      /* ignore partitions */
-      if (udisks_object_peek_partition (object) != NULL)
-        continue;
-
-      if (g_strcmp0 (udisks_block_get_mdraid (block), raid_objpath) == 0)
-        {
-          ret = block;
-          goto out;
-        }
-      g_object_unref (block);
-    }
-
- out:
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
+  b_list = _udisks_client_get_block_or_blocks_for_mdraid (client,
+                                                          raid,
+                                                          udisks_block_get_mdraid,
+                                                          TRUE,   /* Retrieve first one */
+                                                          TRUE);  /* Skip partitions */
+  if (b_list)
+   {
+      ret = (g_list_first (b_list))->data;
+      g_list_free (b_list);
+   }
   return ret;
 }
 
@@ -934,49 +970,11 @@ GList *
 udisks_client_get_all_blocks_for_mdraid (UDisksClient *client,
                                          UDisksMDRaid *raid)
 {
-  GList *ret = NULL;
-  GList *l, *object_proxies = NULL;
-  GDBusObject *raid_object;
-  const gchar *raid_objpath;
-
-  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
-  g_return_val_if_fail (UDISKS_IS_MDRAID (raid), NULL);
-
-  raid_object = g_dbus_interface_get_object (G_DBUS_INTERFACE (raid));
-  if (raid_object == NULL)
-    goto out;
-
-  raid_objpath = g_dbus_object_get_object_path (raid_object);
-
-  object_proxies = g_dbus_object_manager_get_objects (client->object_manager);
-  for (l = object_proxies; l != NULL; l = l->next)
-    {
-      UDisksObject *object = UDISKS_OBJECT (l->data);
-      UDisksBlock *block;
-
-      block = udisks_object_get_block (object);
-      if (block == NULL)
-        continue;
-
-      /* ignore partitions */
-      if (udisks_object_peek_partition (object) != NULL)
-        continue;
-
-      if (g_strcmp0 (udisks_block_get_mdraid (block), raid_objpath) == 0)
-        {
-          ret = g_list_prepend (ret, block);
-        }
-      else
-        {
-          g_object_unref (block);
-        }
-    }
-
- out:
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
-  ret = g_list_reverse (ret);
-  return ret;
+  return g_list_reverse (_udisks_client_get_block_or_blocks_for_mdraid (client,
+                                                                        raid,
+                                                                        udisks_block_get_mdraid,
+                                                                        FALSE,    /* Retrieve all */
+                                                                        TRUE));   /* Skip partitions */
 }
 
 /**
@@ -996,46 +994,12 @@ GList *
 udisks_client_get_members_for_mdraid (UDisksClient *client,
                                       UDisksMDRaid *raid)
 {
-  GList *ret = NULL;
-  GList *l, *object_proxies = NULL;
-  GDBusObject *raid_object;
-  const gchar *raid_objpath;
-
-  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
-  g_return_val_if_fail (UDISKS_IS_MDRAID (raid), NULL);
-
-  raid_object = g_dbus_interface_get_object (G_DBUS_INTERFACE (raid));
-  if (raid_object == NULL)
-    goto out;
-
-  raid_objpath = g_dbus_object_get_object_path (raid_object);
-
-  object_proxies = g_dbus_object_manager_get_objects (client->object_manager);
-  for (l = object_proxies; l != NULL; l = l->next)
-    {
-      UDisksObject *object = UDISKS_OBJECT (l->data);
-      UDisksBlock *block;
-
-      block = udisks_object_get_block (object);
-      if (block == NULL)
-        continue;
-
-      if (g_strcmp0 (udisks_block_get_mdraid_member (block), raid_objpath) == 0)
-        {
-          ret = g_list_prepend (ret, block); /* adopts reference to block */
-        }
-      else
-        {
-          g_object_unref (block);
-        }
-    }
-
- out:
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
-  return ret;
+  return _udisks_client_get_block_or_blocks_for_mdraid (client,
+                                                        raid,
+                                                        udisks_block_get_mdraid_member,
+                                                        FALSE,    /* Retrieve all */
+                                                        FALSE);   /* Don't skip partitions */
 }
-
 /* ---------------------------------------------------------------------------------------------------- */
 
 static void
@@ -1207,8 +1171,7 @@ udisks_client_get_cleartext_block (UDisksClient  *client,
     }
 
  out:
-  g_list_foreach (objects, (GFunc) g_object_unref, NULL);
-  g_list_free (objects);
+  g_list_free_full (objects, g_object_unref);
   return ret;
 }
 
@@ -1262,8 +1225,7 @@ udisks_client_get_drive_siblings  (UDisksClient  *client,
     }
   ret = g_list_reverse (ret);
  out:
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
+  g_list_free_full (object_proxies, g_object_unref);
   return ret;
 }
 
@@ -1314,8 +1276,7 @@ udisks_client_get_partitions (UDisksClient         *client,
     }
   ret = g_list_reverse (ret);
  out:
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
+  g_list_free_full (object_proxies, g_object_unref);
   return ret;
 }
 
@@ -1455,8 +1416,7 @@ udisks_client_get_jobs_for_object (UDisksClient  *client,
     }
   ret = g_list_reverse (ret);
 
-  g_list_foreach (object_proxies, (GFunc) g_object_unref, NULL);
-  g_list_free (object_proxies);
+  g_list_free_full (object_proxies, g_object_unref);
   return ret;
 }
 
@@ -1530,8 +1490,7 @@ on_object_added (GDBusObjectManager  *manager,
     {
       init_interface_proxy (client, G_DBUS_PROXY (l->data));
     }
-  g_list_foreach (interfaces, (GFunc) g_object_unref, NULL);
-  g_list_free (interfaces);
+  g_list_free_full (interfaces, g_object_unref);
 
   udisks_client_queue_changed (client);
 }
@@ -1585,7 +1544,25 @@ on_interface_proxy_properties_changed (GDBusObjectManagerClient   *manager,
                                        gpointer                    user_data)
 {
   UDisksClient *client = UDISKS_CLIENT (user_data);
-  udisks_client_queue_changed (client);
+  UDisksClientClass *client_class = UDISKS_CLIENT_GET_CLASS (client);
+
+  GVariantIter iter;
+  gchar *property_name = NULL;
+
+  /* never emit the change signal for Job objects */
+  if (g_strcmp0 (g_dbus_proxy_get_interface_name (interface_proxy), "org.freedesktop.UDisks2.Drive.Job") == 0)
+    return;
+
+  g_variant_iter_init (&iter, changed_properties);
+  while (g_variant_iter_next (&iter, "{&sv}", &property_name, NULL))
+    {
+      if (! g_hash_table_contains (client_class->changed_blacklist, property_name))
+        {
+          /* one of the properties is not on the blacklist -> emit change signal */
+          udisks_client_queue_changed (client);
+          return;
+        }
+    }
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -1606,27 +1583,31 @@ get_pow2_size (guint64 size)
   gchar *str;
   gdouble displayed_size;
   const gchar *unit;
-  guint digits;
+  gint digits;
 
   if (size < MEBIBYTE_FACTOR)
     {
       displayed_size = (double) size / KIBIBYTE_FACTOR;
-      unit = "KiB";
+      /* Translators: SI prefix and standard unit symbol, translate cautiously (or not at all) */
+      unit = C_("byte-size-pow2", "KiB");
     }
   else if (size < GIBIBYTE_FACTOR)
     {
       displayed_size = (double) size / MEBIBYTE_FACTOR;
-      unit = "MiB";
+      /* Translators: SI prefix and standard unit symbol, translate cautiously (or not at all) */
+      unit = C_("byte-size-pow2", "MiB");
     }
   else if (size < TEBIBYTE_FACTOR)
     {
       displayed_size = (double) size / GIBIBYTE_FACTOR;
-      unit = "GiB";
+      /* Translators: SI prefix and standard unit symbol, translate cautiously (or not at all) */
+      unit = C_("byte-size-pow2", "GiB");
     }
   else
     {
       displayed_size = (double) size / TEBIBYTE_FACTOR;
-      unit = "TiB";
+      /* Translators: SI prefix and standard unit symbol, translate cautiously (or not at all) */
+      unit = C_("byte-size-pow2", "TiB");
     }
 
   if (displayed_size < 10.0)
@@ -1645,27 +1626,31 @@ get_pow10_size (guint64 size)
   gchar *str;
   gdouble displayed_size;
   const gchar *unit;
-  guint digits;
+  gint digits;
 
   if (size < MEGABYTE_FACTOR)
     {
       displayed_size = (double) size / KILOBYTE_FACTOR;
-      unit = "KB";
+      /* Translators: SI prefix and standard unit symbol, translate cautiously (or not at all) */
+      unit = C_("byte-size-pow10", "KB");
     }
   else if (size < GIGABYTE_FACTOR)
     {
       displayed_size = (double) size / MEGABYTE_FACTOR;
-      unit = "MB";
+      /* Translators: SI prefix and standard unit symbol, translate cautiously (or not at all) */
+      unit = C_("byte-size-pow10", "MB");
     }
   else if (size < TERABYTE_FACTOR)
     {
       displayed_size = (double) size / GIGABYTE_FACTOR;
-      unit = "GB";
+      /* Translators: SI prefix and standard unit symbol, translate cautiously (or not at all) */
+      unit = C_("byte-size-pow10", "GB");
     }
   else
     {
       displayed_size = (double) size / TERABYTE_FACTOR;
-      unit = "TB";
+      /* Translators: SI prefix and standard unit symbol, translate cautiously (or not at all) */
+      unit = C_("byte-size-pow10", "TB");
     }
 
   if (displayed_size < 10.0)
@@ -1700,7 +1685,7 @@ udisks_client_get_size_for_display (UDisksClient  *client,
   if (long_string)
     {
       gchar *size_str;
-      size_str = g_strdup_printf ("%'" G_GINT64_FORMAT, size);
+      size_str = g_strdup_printf ("%'" G_GUINT64_FORMAT, size);
       if (use_pow2)
         {
           gchar *pow2_str;
@@ -1904,12 +1889,13 @@ static const struct
   {"filesystem", "vfat",              "FAT12", NC_("fs-type", "FAT (12-bit version)"),              NC_("fs-type", "FAT")},
   {"filesystem", "vfat",              "FAT16", NC_("fs-type", "FAT (16-bit version)"),              NC_("fs-type", "FAT")},
   {"filesystem", "vfat",              "FAT32", NC_("fs-type", "FAT (32-bit version)"),              NC_("fs-type", "FAT")},
-  {"filesystem", "ntfs",              "*",     NC_("fs-type", "FAT (version %s)"),                  NC_("fs-type", "FAT")},
+  {"filesystem", "vfat",              "*",     NC_("fs-type", "FAT (version %s)"),                  NC_("fs-type", "FAT")},
   {"filesystem", "vfat",              NULL,    NC_("fs-type", "FAT"),                               NC_("fs-type", "FAT")},
   {"filesystem", "ntfs",              "*",     NC_("fs-type", "NTFS (version %s)"),                 NC_("fs-type", "NTFS")},
   {"filesystem", "ntfs",              NULL,    NC_("fs-type", "NTFS"),                              NC_("fs-type", "NTFS")},
   {"filesystem", "hfs",               NULL,    NC_("fs-type", "HFS"),                               NC_("fs-type", "HFS")},
   {"filesystem", "hfsplus",           NULL,    NC_("fs-type", "HFS+"),                              NC_("fs-type", "HFS+")},
+  {"filesystem", "apfs",              NULL,    NC_("fs-type", "APFS"),                              NC_("fs-type", "APFS")},
   {"filesystem", "ext2",              "*",     NC_("fs-type", "Ext2 (version %s)"),                 NC_("fs-type", "Ext2")},
   {"filesystem", "ext2",              NULL,    NC_("fs-type", "Ext2"),                              NC_("fs-type", "Ext2")},
   {"filesystem", "ext3",              "*",     NC_("fs-type", "Ext3 (version %s)"),                 NC_("fs-type", "Ext3")},
@@ -1936,10 +1922,12 @@ static const struct
   {"raid",       "linux_raid_member", NULL,    NC_("fs-type", "Linux RAID Member"),                 NC_("fs-type", "Linux RAID Member")},
   {"raid",       "zfs_member",        "*",     NC_("fs-type", "ZFS Device (ZPool version %s)"),     NC_("fs-type", "ZFS (v%s)")},
   {"raid",       "zfs_member",        NULL,    NC_("fs-type", "ZFS Device"),                        NC_("fs-type", "ZFS")},
-  {"raid",       "isw_raid_member",   "*",     NC_("fs-type", "Intel Matrix RAID Member (version %s)"), NC_("fs-type", "IMSM RAID Member (%s)")},
-  {"raid",       "isw_raid_member",   NULL,    NC_("fs-type", "Intel Matrix RAID Member"),          NC_("fs-type", "IMSM RAID")},
+  {"raid",       "isw_raid_member",   "*",     NC_("fs-type", "Intel Rapid Storage Technology enterprise RAID Member (version %s)"), NC_("fs-type", "Intel RSTe RAID Member (%s)")},
+  {"raid",       "isw_raid_member",   NULL,    NC_("fs-type", "Intel Rapid Storage Technology enterprise RAID Member"),          NC_("fs-type", "Intel RSTe RAID Member")},
+  {"crypto",     "crypto_unknown",    NULL,    NC_("fs-type", "Possibly Encrypted"),                NC_("fs-type", "Encrypted?")},
   {"crypto",     "crypto_LUKS",       "*",     NC_("fs-type", "LUKS Encryption (version %s)"),      NC_("fs-type", "LUKS")},
   {"crypto",     "crypto_LUKS",       NULL,    NC_("fs-type", "LUKS Encryption"),                   NC_("fs-type", "LUKS")},
+  {"crypto",     "crypto_TCRYPT",     NULL,    NC_("fs-type", "VeraCrypt/TrueCrypt Encryption"),    NC_("fs-type", "VeraCrypt/TrueCrypt")},
   {"filesystem", "VMFS",              "*",     NC_("fs-type", "VMFS (version %s)"),                 NC_("fs-type", "VMFS (v%s)")},
   {"filesystem", "VMFS",              NULL,    NC_("fs-type", "VMFS"),                              NC_("fs-type", "VMFS")},
   {"raid",       "VMFS_volume_member", "*",    NC_("fs-type", "VMFS Volume Member (version %s)"),   NC_("fs-type", "VMFS Member (v%s)")},
@@ -1967,9 +1955,13 @@ udisks_client_get_id_for_display (UDisksClient *client,
                                   gboolean      long_string)
 {
   guint n;
-  gchar *ret;
+  gchar *ret = NULL;
 
-  ret = NULL;
+  if (usage == NULL || type == NULL || version == NULL)
+    {
+      g_critical ("Invalid device ID specification. Usage, type and version must not be NULL");
+      goto out;
+    }
 
   for (n = 0; id_type[n].usage != NULL; n++)
     {
@@ -1988,8 +1980,8 @@ udisks_client_get_id_for_display (UDisksClient *client,
                    (g_strcmp0 (id_type[n].version, "*") == 0 && strlen (version) > 0))
             {
               /* we know better than the compiler here */
-#ifdef __GNUC_PREREQ
-# if __GNUC_PREREQ(4,6)
+#if defined( __GNUC_PREREQ) || defined(__clang__)
+# if __GNUC_PREREQ(4,6) || __clang__
 #  pragma GCC diagnostic push
 #  pragma GCC diagnostic ignored "-Wformat-nonliteral"
 # endif
@@ -1999,8 +1991,8 @@ udisks_client_get_id_for_display (UDisksClient *client,
               else
                 ret = g_strdup_printf (g_dpgettext2 (GETTEXT_PACKAGE, "fs-type", id_type[n].short_name), version);
               goto out;
-#ifdef __GNUC_PREREQ
-# if __GNUC_PREREQ(4,6)
+#if defined( __GNUC_PREREQ) || defined(__clang__)
+# if __GNUC_PREREQ(4,6) || __clang__
 #  pragma GCC diagnostic pop
 # endif
 #endif
@@ -2210,10 +2202,19 @@ static const struct
   {"gpt", "generic",   "024dee41-33e7-11d3-9d69-0008c781f39f", NC_("part-type", "MBR Partition Scheme"), F_SYSTEM},
   {"gpt", "generic",   "c12a7328-f81f-11d2-ba4b-00a0c93ec93b", NC_("part-type", "EFI System"), F_SYSTEM},
   {"gpt", "generic",   "21686148-6449-6e6f-744e-656564454649", NC_("part-type", "BIOS Boot"), F_SYSTEM},
+  /* This is also defined in the Apple and Solaris section */
+  {"gpt", "generic",   "6a898cc3-1dd2-11b2-99a6-080020736631", NC_("part-type", "ZFS"), 0},
+  /* Extended Boot Partition, see http://www.freedesktop.org/wiki/Specifications/BootLoaderSpec/ */
+  {"gpt", "generic",   "bc13c2ff-59e6-4262-a352-b275fd6f7172", NC_("part-type", "Extended Boot Partition"), 0},
+  /* Discoverable Linux Partitions, see http://www.freedesktop.org/wiki/Specifications/DiscoverablePartitionsSpec */
+  {"gpt", "linux",     "44479540-f297-41b2-9af7-d131d5f0458a", NC_("part-type", "Linux Root Partition (x86)"), 0},
+  {"gpt", "linux",     "4f68bce3-e8cd-4db1-96e7-fbcaf984b709", NC_("part-type", "Linux Root Partition (x86_64)"), 0},
+  {"gpt", "linux",     "933ac7e1-2eb4-4f13-b844-0e14e2aef915", NC_("part-type", "Linux Home Partition"), 0},
+  {"gpt", "linux",     "3b8f8425-20e0-4f3b-907f-1a25a76f98e8", NC_("part-type", "Linux Server Data Partition"), 0},
   /* Linux */
+  {"gpt", "linux",     "0657fd6d-a4ab-43c4-84e5-0933c84b4f4f", NC_("part-type", "Linux Swap"), F_SWAP},
   {"gpt", "linux",     "0fc63daf-8483-4772-8e79-3d69d8477de4", NC_("part-type", "Linux Filesystem"), 0},
   {"gpt", "linux",     "a19d880f-05fc-4d3b-a006-743f0f84911e", NC_("part-type", "Linux RAID"), F_RAID},
-  {"gpt", "linux",     "0657fd6d-a4ab-43c4-84e5-0933c84b4f4f", NC_("part-type", "Linux Swap"), F_SWAP},
   {"gpt", "linux",     "e6d6d379-f507-44c2-a23c-238f2a3df928", NC_("part-type", "Linux LVM"), F_RAID},
   {"gpt", "linux",     "8da63339-0007-60c0-c436-083ac8230908", NC_("part-type", "Linux Reserved"), 0},
   /* Microsoft */
@@ -2224,8 +2225,9 @@ static const struct
   {"gpt", "microsoft", "de94bba4-06d1-4d40-a16a-bfd50179d6ac", NC_("part-type", "Microsoft Windows Recovery Environment"), 0},
   /* Apple OS X */
   {"gpt", "apple",     "48465300-0000-11aa-aa11-00306543ecac", NC_("part-type", "Apple HFS/HFS+"), 0},
+  {"gpt", "apple",     "7c3457ef-0000-11aa-aa11-00306543ecac", NC_("part-type", "Apple APFS"), 0},
   {"gpt", "apple",     "55465300-0000-11aa-aa11-00306543ecac", NC_("part-type", "Apple UFS"), 0},
-  {"gpt", "apple",     "6a898cc3-1dd2-11b2-99a6-080020736631", NC_("part-type", "Apple ZFS"), 0}, /* same as Solaris /usr */
+  {"gpt", "apple",     "6a898cc3-1dd2-11b2-99a6-080020736631", NC_("part-type", "Apple ZFS"), 0}, /* same as ZFS */
   {"gpt", "apple",     "52414944-0000-11aa-aa11-00306543ecac", NC_("part-type", "Apple RAID"), F_RAID},
   {"gpt", "apple",     "52414944-5f4f-11aa-aa11-00306543ecac", NC_("part-type", "Apple RAID (offline)"), F_RAID},
   {"gpt", "apple",     "426f6f74-0000-11aa-aa11-00306543ecac", NC_("part-type", "Apple Boot"), F_SYSTEM},
@@ -2247,7 +2249,7 @@ static const struct
   {"gpt", "other",     "6a85cf4d-1dd2-11b2-99a6-080020736631", NC_("part-type", "Solaris Root"), 0},
   {"gpt", "other",     "6a87c46f-1dd2-11b2-99a6-080020736631", NC_("part-type", "Solaris Swap"), F_SWAP},
   {"gpt", "other",     "6a8b642b-1dd2-11b2-99a6-080020736631", NC_("part-type", "Solaris Backup"), 0},
-  {"gpt", "other",     "6a898cc3-1dd2-11b2-99a6-080020736631", NC_("part-type", "Solaris /usr"), 0}, /* same as Apple ZFS */
+  {"gpt", "other",     "6a898cc3-1dd2-11b2-99a6-080020736631", NC_("part-type", "Solaris /usr"), 0}, /* same as ZFS */
   {"gpt", "other",     "6a8ef2e9-1dd2-11b2-99a6-080020736631", NC_("part-type", "Solaris /var"), 0},
   {"gpt", "other",     "6a90ba39-1dd2-11b2-99a6-080020736631", NC_("part-type", "Solaris /home"), 0},
   {"gpt", "other",     "6a9283a5-1dd2-11b2-99a6-080020736631", NC_("part-type", "Solaris Alternate Sector"), 0},
@@ -2266,6 +2268,15 @@ static const struct
   /* VMWare, see http://blogs.vmware.com/vsphere/2011/08/vsphere-50-storage-features-part-7-gpt.html */
   {"gpt", "other",     "aa31e02a-400f-11db-9590-000c2911d1b8", NC_("part-type", "VMWare VMFS"), 0},
   {"gpt", "other",     "9d275380-40ad-11db-bf97-000c2911d1b8", NC_("part-type", "VMWare vmkcore"), 0},
+  /* ChromeOS, see http://www.chromium.org/chromium-os/chromiumos-design-docs/disk-format */
+  {"gpt", "other",     "cab6e88e-abf3-4102-a07a-d4bb9be3c1d3", NC_("part-type", "ChromeOS Firmware"), 0},
+  {"gpt", "other",     "fe3a2a5d-4f32-41a7-b725-accc3285a309", NC_("part-type", "ChromeOS Kernel"), 0},
+  {"gpt", "other",     "3cb8e202-3b7e-47dd-8a3c-7ff2a13cfcec", NC_("part-type", "ChromeOS Root Filesystem"), 0},
+  {"gpt", "other",     "2e0a753d-9e48-43b0-8337-b15192cb1b5e", NC_("part-type", "ChromeOS Reserved"), 0},
+  /* Intel Partition Types */
+  /*     FFS = Fast Flash Standby, aka Intel Rapid start  */
+  /*     http://downloadmirror.intel.com/22647/eng/Intel%20Rapid%20Start%20Technology%20Deployment%20Guide%20v1.0.pdf */
+  {"gpt", "other",     "d3bfe2de-3daf-11df-ba40-e3a556d89593", NC_("part-type", "Intel FFS Reserved"), 0},
 
   /* see http://developer.apple.com/documentation/mac/devices/devices-126.html
    *     http://lists.apple.com/archives/Darwin-drivers/2003/May/msg00021.html */
@@ -2295,7 +2306,7 @@ static const struct
   {"dos", "microsoft", "0x01",  NC_("part-type", "FAT12"), 0},
   {"dos", "microsoft", "0x04",  NC_("part-type", "FAT16 <32M"), 0},
   {"dos", "microsoft", "0x06",  NC_("part-type", "FAT16"), 0},
-  {"dos", "microsoft", "0x07",  NC_("part-type", "HPFS/NTFS"), 0},
+  {"dos", "microsoft", "0x07",  NC_("part-type", "NTFS/exFAT/HPFS"), 0},
   {"dos", "microsoft", "0x0b",  NC_("part-type", "W95 FAT32"), 0},
   {"dos", "microsoft", "0x0c",  NC_("part-type", "W95 FAT32 (LBA)"), 0},
   {"dos", "microsoft", "0x0e",  NC_("part-type", "W95 FAT16 (LBA)"), 0},
@@ -2394,33 +2405,70 @@ udisks_client_get_partition_type_for_display (UDisksClient  *client,
   return ret;
 }
 
+/**
+ * udisks_client_get_partition_type_and_subtype_for_display:
+ * @client: A #UDisksClient.
+ * @partition_table_type: A partitioning type e.g. 'dos' or 'gpt'.
+ * @partition_table_subtype: A partitioning subtype or %NULL.
+ * @partition_type: A partition type.
+ *
+ * Like udisks_client_get_partition_type_for_display() but also takes
+ * the partition table subtype into account, if available. This is
+ * useful in scenarios where different subtypes is using the same
+ * partition type.
+ *
+ * Returns: A description of @partition_type or %NULL if unknown.
+ *
+ * Since: 2.1.1
+ */
+const gchar *
+udisks_client_get_partition_type_and_subtype_for_display (UDisksClient  *client,
+                                                          const gchar   *partition_table_type,
+                                                          const gchar   *partition_table_subtype,
+                                                          const gchar   *partition_type)
+{
+  const gchar *ret = NULL;
+  guint n;
+
+  for (n = 0; known_partition_types[n].name != NULL; n++)
+    {
+      if (g_strcmp0 (known_partition_types[n].table_type, partition_table_type) == 0 &&
+          g_strcmp0 (known_partition_types[n].type, partition_type) == 0)
+        {
+          if (partition_table_subtype != NULL &&
+              g_strcmp0 (known_partition_types[n].table_subtype, partition_table_subtype) != 0)
+            continue;
+          ret = g_dpgettext2 (GETTEXT_PACKAGE, "part-type", known_partition_types[n].name);
+          goto out;
+        }
+    }
+
+ out:
+  return ret;
+}
+
 /* ---------------------------------------------------------------------------------------------------- */
 
 /**
- * udisks_client_get_job_description:
- * @client: A #UDisksClient.
- * @job: A #UDisksJob.
+ * udisks_client_get_job_description_from_operation:
+ * @operation: A job operation name.
  *
- * Gets a human-readable and localized text string describing the
- * operation of @job.
+ * Gets a human-readable and localized text string describing a
+ * a job @operation.
  *
- * For known job types, see the documentation for the
+ * For known job operation types, see the documentation for the
  * <link linkend="gdbus-property-org-freedesktop-UDisks2-Job.Operation">Job:Operation</link>
  * D-Bus property.
  *
  * Returns: A string that should be freed with g_free().
  */
 gchar *
-udisks_client_get_job_description (UDisksClient   *client,
-                                   UDisksJob      *job)
+udisks_client_get_job_description_from_operation (const gchar *operation)
 {
   static gsize once = 0;
   static GHashTable *hash = NULL;
-  const gchar *operation = NULL;
   gchar *ret = NULL;
 
-  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
-
   if (g_once_init_enter (&once))
     {
       hash = g_hash_table_new (g_str_hash, g_str_equal);
@@ -2429,11 +2477,16 @@ udisks_client_get_job_description (UDisksClient   *client,
       g_hash_table_insert (hash, (gpointer) "encrypted-unlock",     (gpointer) C_("job", "Unlocking Device"));
       g_hash_table_insert (hash, (gpointer) "encrypted-lock",       (gpointer) C_("job", "Locking Device"));
       g_hash_table_insert (hash, (gpointer) "encrypted-modify",     (gpointer) C_("job", "Modifying Encrypted Device"));
+      g_hash_table_insert (hash, (gpointer) "encrypted-resize",     (gpointer) C_("job", "Resizing Encrypted Device"));
       g_hash_table_insert (hash, (gpointer) "swapspace-start",      (gpointer) C_("job", "Starting Swap Device"));
       g_hash_table_insert (hash, (gpointer) "swapspace-stop",       (gpointer) C_("job", "Stopping Swap Device"));
+      g_hash_table_insert (hash, (gpointer) "swapspace-modify",     (gpointer) C_("job", "Modifying Swap Device"));
+      g_hash_table_insert (hash, (gpointer) "filesystem-check",     (gpointer) C_("job", "Checking Filesystem"));
       g_hash_table_insert (hash, (gpointer) "filesystem-mount",     (gpointer) C_("job", "Mounting Filesystem"));
       g_hash_table_insert (hash, (gpointer) "filesystem-unmount",   (gpointer) C_("job", "Unmounting Filesystem"));
       g_hash_table_insert (hash, (gpointer) "filesystem-modify",    (gpointer) C_("job", "Modifying Filesystem"));
+      g_hash_table_insert (hash, (gpointer) "filesystem-repair",    (gpointer) C_("job", "Repairing Filesystem"));
+      g_hash_table_insert (hash, (gpointer) "filesystem-resize",    (gpointer) C_("job", "Resizing Filesystem"));
       g_hash_table_insert (hash, (gpointer) "format-erase",         (gpointer) C_("job", "Erasing Device"));
       g_hash_table_insert (hash, (gpointer) "format-mkfs",          (gpointer) C_("job", "Creating Filesystem"));
       g_hash_table_insert (hash, (gpointer) "loop-setup",           (gpointer) C_("job", "Setting Up Loop Device"));
@@ -2453,15 +2506,37 @@ udisks_client_get_job_description (UDisksClient   *client,
       g_once_init_leave (&once, (gsize) 1);
     }
 
-  operation = udisks_job_get_operation (job);
   if (operation != NULL)
     ret = g_strdup (g_hash_table_lookup (hash, operation));
   if (ret == NULL)
-    ret = g_strdup_printf (C_("unknown-job", "Unknown (%s)"), udisks_job_get_operation (job));
+    ret = g_strdup_printf (C_("unknown-job", "Unknown (%s)"), operation == NULL ? "(null)" : operation);
 
   return ret;
 }
 
+/**
+ * udisks_client_get_job_description:
+ * @client: A #UDisksClient.
+ * @job: A #UDisksJob.
+ *
+ * Gets a human-readable and localized text string describing the
+ * operation of @job.
+ *
+ * For known job types, see the documentation for the
+ * <link linkend="gdbus-property-org-freedesktop-UDisks2-Job.Operation">Job:Operation</link>
+ * D-Bus property.
+ *
+ * Returns: A string that should be freed with g_free().
+ */
+gchar *
+udisks_client_get_job_description (UDisksClient   *client,
+                                   UDisksJob      *job)
+{
+  g_return_val_if_fail (UDISKS_IS_CLIENT (client), NULL);
+
+  return udisks_client_get_job_description_from_operation (udisks_job_get_operation (job));
+}
+
 /* ---------------------------------------------------------------------------------------------------- */
 
 static UDisksPartitionTypeInfo *
