@@ -259,7 +259,7 @@ udisks_client_get_object_info_for_block (UDisksClient     *client,
   if (partition != NULL)
     {
       /* Translators: Used to describe a partition of a block device.
-       *              The %d is the partition number.
+       *              The %u is the partition number.
        *              The %s is the description for the block device (e.g. "5 GB Block Device").
        */
       s = g_strdup_printf (C_("part-block", "Partition %u of %s"),
@@ -279,6 +279,8 @@ udisks_client_get_object_info_for_block (UDisksClient     *client,
   info->sort_key = g_strdup_printf ("02_block_%s_%u",
                                     last_segment (g_dbus_object_get_object_path (G_DBUS_OBJECT (info->object))),
                                     partition != NULL ? udisks_partition_get_number (partition) : 0);
+
+  g_free (size_str);
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -313,7 +315,7 @@ udisks_client_get_object_info_for_loop (UDisksClient     *client,
   if (partition != NULL)
     {
       /* Translators: Used to describe a partition of a loop device.
-       *              The %d is the partition number.
+       *              The %u is the partition number.
        *              The %s is the description for the loop device (e.g. "5 GB Loop Device").
        */
       s = g_strdup_printf (C_("part-loop", "Partition %u of %s"),
@@ -335,6 +337,8 @@ udisks_client_get_object_info_for_loop (UDisksClient     *client,
   info->sort_key = g_strdup_printf ("03_loop_%s_%u",
                                     last_segment (g_dbus_object_get_object_path (G_DBUS_OBJECT (info->object))),
                                     partition != NULL ? udisks_partition_get_number (partition) : 0);
+
+  g_free (size_str);
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -409,7 +413,7 @@ udisks_client_get_object_info_for_mdraid (UDisksClient     *client,
   if (partition != NULL)
     {
       /* Translators: Used to describe a partition of a RAID Array.
-       *              The %d is the partition number.
+       *              The %u is the partition number.
        *              The %s is the description for the drive (e.g. "2 TB RAID-5").
        */
       s = g_strdup_printf (C_("part-raid", "Partition %u of %s"),
@@ -471,6 +475,8 @@ udisks_client_get_object_info_for_mdraid (UDisksClient     *client,
 
   info->sort_key = g_strdup_printf ("01_mdraid_%s_%u", udisks_mdraid_get_uuid (mdraid),
                                     partition != NULL ? udisks_partition_get_number (partition) : 0);
+
+  g_free (size_str);
 }
 
 /* ---------------------------------------------------------------------------------------------------- */
@@ -826,7 +832,7 @@ udisks_client_get_object_info_for_drive (UDisksClient     *client,
   if (partition != NULL)
     {
       /* Translators: Used to describe a partition of a drive.
-       *              The %d is the partition number.
+       *              The %u is the partition number.
        *              The %s is the description for the drive (e.g. "2 GB Thumb Drive").
        */
       s = g_strdup_printf (C_("part-drive", "Partition %u of %s"),
